#!/bin/bash
################################################################################
# redirectModbus.sh
#
# Author: Piotr Golonka, CERN EN/ICE-SCD
# Date: 10 February 2012
#
# Usage:
#  redirectModubus.sh <PLC_IP> <PVSS00mod_port>
#
# Description:
# ------------
# This script reconfigures the Linux IP stack (iptables) to reroute the 
# packets for the PVSS MODBUS driver (port 502), so that many projects could run
# on the same machine.
#
# The scripts allow to configure the redirection of packets depending on the IP 
# address they are coming from. The packets are redirected to the same machine, 
# yet to a specified port.
#
# Such solution solves a problem of MODBUS in "push" mode, where the destination 
# TCP/IP port needs to be equal to 502. If we want to run multiple PVSS projects
# in the same machine, we need more instances of the MODBUS driver running,
# and each of them needs to listen on a different TCP/IP port. 
#
# The script needs to be executed by root or by a user with sudo rights
# (in the latter case, sudo may prompt for user's password).
#
# Note that the script is intelligent enough to see whether the redirection
# rule already exists for a specified PLC, or specified redirected port, and
# adapts the settings appropriately
#
# Example: 
# --------
#  PLC1 (172.18.25.136) should be connected to PVSS00mod of Project1
#  PLC2 (172.18.37.9) should be connected to PVSS00mod of Project2
#
#  - configure PVSS00mod on Project1 to start with port number 8502
#   and on Project2 with port number 9502. You do it in the config file, eg
#   for Project1:
#	[mod_2]
#	tcpServerPort=8502
#
#  - to activate the redirection, execute
#	redirectModbus.sh 172.18.25.136 8502
#	redirectModbus.sh 172.18.37.9   9502
#
# For more details about the origins, look, for instance, at:
# https://icecontrols.cern.ch/jira/browse/ENS-1922
# https://icecontrols.cern.ch/jira/browse/ENS-3875
################################################################################

if [ $# != 2 ]; then
 echo "Usage: $0 <PLC_IP> <PVSS00mod_port>"
 exit -1;
fi

PLC_IP=$1
DEST_PORT=$2

# guess what is the IP address on which the packet comes by looking up the route to this IP
MY_IP=`/sbin/ip route get $PLC_IP |awk '/src/{print $7;}'`


## Check if an entry already exist in the iptables
OUT=`sudo /sbin/iptables -t nat -n --line-numbers -L PREROUTING |awk -v plc_ip=${PLC_IP} -v dport=${DEST_PORT} '/DNAT/{split($9,a,"(.)*:"); if ($5==plc_ip || a[2]==dport) print $1,$5,a[2]};'`

if [ -z "${OUT}" ]; then
    # it does not exist in the table... add it
    sudo /sbin/iptables -t nat -A PREROUTING -p tcp -s ${PLC_IP} --dport 502 -d ${MY_IP} -j DNAT --to-destination $MY_IP:${DEST_PORT}     
else
    # otherwise, replace the existing rule
    RULE_TO_REPLACE=`echo $OUT|awk '{print $1}'`;
    sudo /sbin/iptables -t nat -R PREROUTING ${RULE_TO_REPLACE} -p tcp -s ${PLC_IP} --dport 502 -d ${MY_IP} -j DNAT --to-destination $MY_IP:${DEST_PORT} 
fi;
