/**@file

// libPackageCreationFrontEndTypeFunctions.ctl
This library contains functions used to manipulate front-end types in the package creation application.

@par Creation Date
  30/11/2011

@par Modification History
  30/11/2011 - First release
  
@par Constraints
  Can only be used by the main panel of the package creation application (dependencies to variables and widgets of the panel)

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@author 
  Cyril Caillaba (EN-ICE)
*/

//@{


// ------------------------------------------------------------------------------------ //
// -----------------libPackageCreationFrontEndType_addNewFrontEndType------------------ //
// ------------------------------------------------------------------------------------ //
/** Adds a new front-end type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sFrontEndTypeName  input, name of the front-end type to add to the list.
@param bAlreadyExists input, whether the front-end already exists somewhere else or not.
@param bBefore intput, true to insert before the currently selected front-end, false to insert after.
@param dsExceptions output, the list of exceptions that occur in the function.
*/
void libPackageCreationFrontEndType_addNewFrontEndType(
    string sFrontEndTypeName,
    bool bAlreadyExists,
    bool bBefore,
	dyn_string &dsExceptions)
{
  
  // Check if the name is empty
  if (sFrontEndTypeName == "")
  {
    string sError = getCatStr("unPackageGeneration","EDIT-EMPTYFRONTENDNAME");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-EMPTYFRONTENDNAME"
    ); 
    return; 
  }
  
  // Always replace spaces by underscores
  strreplace(sFrontEndTypeName, " ", "_");
  
  shape tableFrontEnd = getShape("tableFrontEndList"); 
  
  // Check if this front-end type is not already in the list
  for (int i = 0 ; i < tableFrontEnd.lineCount() ; i++)
  {
    string currentFrontEndTypeName;
    getValue("tableFrontEndList","cellValueRC",i,COLUMN_FRONT_END_NAME,currentFrontEndTypeName);  
    if (currentFrontEndTypeName == sFrontEndTypeName)
    {
      string sError = getCatStr("unPackageGeneration","EDIT-FRONTENDALREADYEXISTS");
      strreplace(sError,"FRONTEND",sFrontEndTypeName);
      
      fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-FRONTENDALREADYEXISTS"
      ); 
      return; 
    }
      
  }    
 
  if (currentCellFrontEnd > -1)
  {
    if (bBefore)
    {
      tableFrontEnd.insertLineN(currentCellFrontEnd);
      tableFrontEnd.cellValueRC(currentCellFrontEnd,COLUMN_FRONT_END_NAME,sFrontEndTypeName);
      tableFrontEnd.cellValueRC(currentCellFrontEnd,COLUMN_FRONT_END_EXISTS,bAlreadyExists);
    }
    else
    {
      tableFrontEnd.insertLineN(currentCellFrontEnd+1);
      tableFrontEnd.cellValueRC(currentCellFrontEnd+1,COLUMN_FRONT_END_NAME,sFrontEndTypeName);
      tableFrontEnd.cellValueRC(currentCellFrontEnd+1,COLUMN_FRONT_END_EXISTS,bAlreadyExists);
    }
    
  }
  else
  {
    if (bBefore)
    {
      tableFrontEnd.insertLineN(0);
      tableFrontEnd.cellValueRC(0,COLUMN_FRONT_END_NAME,sFrontEndTypeName);
      tableFrontEnd.cellValueRC(0,COLUMN_FRONT_END_EXISTS,bAlreadyExists);
    }
    else 
    {
      tableFrontEnd.appendLine(COLUMN_FRONT_END_NAME, sFrontEndTypeName, COLUMN_FRONT_END_EXISTS, bAlreadyExists);
    }
  }

  // Clear the selected line, so that we add at the end for the next front-end
  currentCellFrontEnd = -1;
  
  // Clear the content of the text field
  textFieldNewFrontEndType.text("");
}

// ------------------------------------------------------------------------------------ //
// ------------------libPackageCreationFrontEndType_deleteFrontEndType----------------- //
// ------------------------------------------------------------------------------------ //
/** Deletes the currently selected front-end type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void libPackageCreationFrontEndType_deleteFrontEndType()
{
  if (currentCellFrontEnd == -1)
  {
    return;
  } 
  else
  {
    setValue("tableFrontEndList","deleteLineN",currentCellFrontEnd);
    currentCellFrontEnd = -1;
  }
}


// ------------------------------------------------------------------------------------ //
// --------------libPackageCreationFrontEndType_deleteAllFrontEndTypes----------------- //
// ------------------------------------------------------------------------------------ //

/** Deletes all the front-end types of the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void libPackageCreationFrontEndType_deleteAllFrontEndTypes()
{
  int lineCount;
  getValue("tableFrontEndList","lineCount",lineCount);
  setValue("tableFrontEndList","deleteLinesN",0,lineCount);
  currentCellFrontEnd = -1;
}

// ------------------------------------------------------------------------------------ //
// -------------------libPackageCreationFrontEndType_frontEndExists-------------------- //
// ------------------------------------------------------------------------------------ //
/** Checks if a front-end exists. Exists means the matching checkbox is checked.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sFrontEndName  input, name of the front-end type to check existance.
@param bExists output, whether the front-end is checked as existing or not.
@return -1 if the front-end type is not found in the list, 0 if everything goes well.
*/
int libPackageCreationFrontEndType_frontEndExists(string sFrontEndName, bool &bExists)
{
  bExists = false;
  for (int i = 1 ; i <= dynlen(frontEndTypeList) ; i++)
  {
    if (frontEndTypeList[i][LIST_NAME_INDEX] == sFrontEndName)
    {
      bExists = frontEndTypeList[i][LIST_EXISTS_INDEX];
      return 0;    
    }
  }
  
  return -1;
}


// -------------------------------------------------------------------------------------- //
// ------------------libPackageCreationFrontEndType_createNewFrontEnd-------------------- //
// -------------------------------------------------------------------------------------- //
/** Creates all necessary information for a new front-end.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sFrontEndType  input, the name of the front-end type to create.
@param mFileFunctions output, each key correspond to a file that will need to be created, and the value is a list of elements inside this file (functions mainly)
@param sTemplateSourcePath  input, The path where the template is located.
@param sPrefix  input, the prefix used for the generation.
@param sPackage  input, the name of the package.
@param sNewPackagePath  input, the path where to save the new package data.
@return -1 in case of failure, 0 else.
*/
int libPackageCreationFrontEndType_createNewFrontEnd(string sFrontEndType, mapping &mFileFunctions, string sTemplateSourcePath, string sPrefix, string sPackage, string sNewPackagePath)
{
  
  
  // ------------------------------------------------------------------ STEP 1 : DPT creation ------------------------------------------------------------------------//
  
  
  // Create the DPT also means exporting it to ASCII to be imported later.
  // So let's just generate an ASCII dump directly  
  // -------------------------------------- //
  // ----- File dplist/TAG_package.dpl ---- //
  // -------------------------------------- //  
  
  if (!mappingHasKey(mFileFunctions,"dplist/" + sPackage + ".dpl"))
  {
    mFileFunctions["dplist/" + sPackage + ".dpl"] = makeDynString();
    dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],"# ascii dump of database\n\n");  
  }
  
  
  string dpTypeFrontEnd = 
      "\n\n# DpType\n"
      "TypeName\n" +
      sFrontEndType + "." + sFrontEndType + "\t" + DPEL_STRUCT + "#1\n"
        "\tconfiguration\t"       + DPEL_STRUCT     + "#2\n"
          "\t\timportTime\t"      + DPEL_DYN_TIME   + "#3\n"  
          "\t\tarchive_bool\t"    + DPEL_DYN_STRING + "#4\n"
          "\t\tarchive_analog\t"  + DPEL_DYN_STRING + "#5\n"
          "\t\tarchive_event\t"   + DPEL_DYN_STRING + "#6\n"
          "\t\tsubApplications\t" + DPEL_DYN_STRING + "#7\n"
        "\tcommunication\t"       + DPEL_STRUCT     + "#8\n"
          "\t\tdriver_num\t"      + DPEL_INT        + "#9\n"
          "\t\tcounter\t"         + DPEL_INT        + "#10\n"
      
      ;
  
  
  dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],dpTypeFrontEnd);  
  
  
  // ------------------------------------------------------------------ STEP 2 : DP front-end definition ------------------------------------------------------------------ //
    
  
  // Generation of ASCII dump
  unsigned newDpId;
  int dpeId;
  
  // Creation of a DP of type _UnFrontEnd
  // If it already exists the old one is removed
  if (dpExists(sFrontEndType+"_unicosInfo"))
  {  
    dpDelete(sFrontEndType+"_unicosInfo");
  }
  dpCreate(sFrontEndType+"_unicosInfo","_UnFrontEnd");  
  dpGetId(sFrontEndType+"_unicosInfo", newDpId, dpeId);
  
  string dpFrontEnd = 
      "\n\n# Datapoint/DpId\n"
      "DpName	TypeName	ID\n" +
      sFrontEndType+"_unicosInfo\t_UnFrontEnd\t" + newDpId + "\n";
  
    
  // Creation of a DP of type _UnSystemIntegrity
  // If it already exists the old one is removed
  if (dpExists(sFrontEndType+"_systemIntegrityInfo"))
  {  
    dpDelete(sFrontEndType+"_systemIntegrityInfo");
  }
  dpCreate(sFrontEndType+"_systemIntegrityInfo","_UnSystemIntegrity");  
  dpGetId(sFrontEndType+"_systemIntegrityInfo", newDpId, dpeId);
  
  dpFrontEnd += 
      sFrontEndType+"_systemIntegrityInfo\t_UnSystemIntegrity\t" + newDpId + "\n";

  
  // Creation of a DP of type _UnTrendConfiguration
  // If it already exists the old one is removed
  if (dpExists(sFrontEndType+"_trendConfiguration"))
  {  
    dpDelete(sFrontEndType+"_trendConfiguration");
  }
  dpCreate(sFrontEndType+"_trendConfiguration","_UnTrendConfiguration");  
  dpGetId(sFrontEndType+"_trendConfiguration", newDpId, dpeId);  
  
  dpFrontEnd += 
      sFrontEndType+"_trendConfiguration\t_UnTrendConfiguration\t" + newDpId + "\n";

  
  // Creation of a DP of type FwTrendingPlot
  // If it already exists the old one is removed
  if (dpExists("_template_un" + sFrontEndType + "_Trend"))
  {  
    dpDelete("_template_un" + sFrontEndType + "_Trend");
  }
  dpCreate("_template_un" + sFrontEndType + "_Trend","FwTrendingPlot");  
  dpGetId("_template_un" + sFrontEndType + "_Trend", newDpId, dpeId);  
  
  dpFrontEnd += 
      "_template_un" + sFrontEndType + "_Trend\tFwTrendingPlot\t211469\n";
  
  
  
  
  dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],dpFrontEnd);  
  
  
  // Generating dp values dump    
  string stringFileTAG_package_dpl;
  fileToString(sTemplateSourcePath + "/dplist/TAG_package.dpl",stringFileTAG_package_dpl);

  // Extract dpValues section
  string dpValuesSection = substr(stringFileTAG_package_dpl,strpos(stringFileTAG_package_dpl,"# DpValue") + 10); // +10 to skip the comment
  
  // Separate each line
  dyn_string dpValuesLines = strsplit(dpValuesSection,"\n");
  string dpValues = "\n\n# DpValue\n" + dpValuesLines[1] +"\n";
  
  for (int i = 2 ; i <= dynlen(dpValuesLines) ; i++)
  {
    string currentLine = dpValuesLines[i];
    if (strpos(currentLine,"TAG_FrontEndType") >= 0)
    {
      // The line concerns a front-end, it is treated
      // TAG_* replacements
      
      
      // Date replacement
      // Date is at the end of the line, e.g. 30.11.2009 11:51:22.399 (23 last characters)
      string currentLineMinusDate = substr(currentLine, 0, strlen(currentLine) - 23);
      time now = getCurrentTime();
      currentLineMinusDate += formatTime("%d.%m.%Y %H:%M:%S.",now,"%03d");
      strreplace(currentLineMinusDate,"TAG_FrontEndType",sFrontEndType);
      strreplace(currentLineMinusDate,"TAG_prefix",sPrefix);
      strreplace(currentLineMinusDate,"TAG_PACKAGE",strtoupper(sPackage));
      
      dpValues += currentLineMinusDate + "\n";
    }
    // Else it is ignored
    
  }
  
  
  dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],dpValues);  

  
  // ------------------------------------------------------------------ STEP 3 : check/delete/import ------------------------------------------------------------------ //   
  
  // Simply open both front-end files and replace the tags in every function
  
  
  
  // ----------------------------------------------------------------- //
  // ----- File scripts/libs/TAG_prefixConfigTAG_FrontEndType.ctl ---- //
  // ----------------------------------------------------------------- //  
  string stringFileTAG_prefixConfigTAG_FrontEndType;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixConfigTAG_FrontEndType.ctl",stringFileTAG_prefixConfigTAG_FrontEndType);

  mFileFunctions["scripts/libs/" + sPrefix + "Config" + sFrontEndType + ".ctl"] = makeDynString();
  
  // Replacement is proceeded each function to have a better control of the process and possible future changes
  dyn_string functionListTAG_prefixConfigTAG_FrontEndType = makeDynString();
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_checkDeleteCommand");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_checkFrontEnd");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_deleteCommand");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_setFrontEnd");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_checkApplication");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_checkData");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_checkAdditionalData");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_setApplication");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_prefixConfigTAG_FrontEndType_deleteFrontEnd");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_FrontEndType_checkFESystemAlarm");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_FrontEndType_setFESystemAlarm");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_FrontEndType_getFrontEndArchiveDp");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_FrontEndType_getArchiveProxyDPE");
  dynAppend(functionListTAG_prefixConfigTAG_FrontEndType,"TAG_FrontEndType_getFrontEndManagerConfig");
  
  for (int i = 1 ; i <= dynlen(functionListTAG_prefixConfigTAG_FrontEndType) ; i++)
  {
    string functionContent = extractFunctionFromStringFile(stringFileTAG_prefixConfigTAG_FrontEndType,functionListTAG_prefixConfigTAG_FrontEndType[i]);
    // Tag replacement
    strreplace(functionContent,"TAG_prefix",sPrefix); 
    strreplace(functionContent,"TAG_FrontEndType",sFrontEndType); 
    strreplace(functionContent,"TAG_FRONTENDTYPE",strtoupper(sFrontEndType)); 
    strreplace(functionContent,"TAG_PREFIX",strtoupper(sPrefix)); 
    
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "Config" + sFrontEndType + ".ctl"],functionContent);  
  }
  
  
  // ----------------------------------------------------------------- //
  // -------- File scripts/libs/TAG_prefixTAG_FrontEndType.ctl ------- //
  // ----------------------------------------------------------------- //  
  string stringFileTAG_prefixTAG_FrontEndType;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixTAG_FrontEndType.ctl",stringFileTAG_prefixTAG_FrontEndType);

  mFileFunctions["scripts/libs/" + sPrefix + sFrontEndType + ".ctl"] = makeDynString();
  
  // Replacement is proceeded each function to have a better control of the process and possible future changes
  dyn_string functionListTAG_prefixTAG_FrontEndType = makeDynString();
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_ObjectListGetValueTime");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_MenuConfiguration");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_HandleMenu");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_WidgetRegisterCB");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_WidgetDisconnection");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_WidgetConnect");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_WidgetDisconnect");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_WidgetCB");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_FaceplateRegisterCB");
  dynAppend(functionListTAG_prefixTAG_FrontEndType,"TAG_FrontEndType_FaceplateDisconnection");
  
  for (int i = 1 ; i <= dynlen(functionListTAG_prefixTAG_FrontEndType) ; i++)
  {
    string functionContent = extractFunctionFromStringFile(stringFileTAG_prefixTAG_FrontEndType,functionListTAG_prefixTAG_FrontEndType[i]);
    // Tag replacement
    strreplace(functionContent,"TAG_prefix",sPrefix); 
    strreplace(functionContent,"TAG_FrontEndType",sFrontEndType); 
    strreplace(functionContent,"TAG_FRONTENDTYPE",strtoupper(sFrontEndType)); 
    strreplace(functionContent,"TAG_PREFIX",strtoupper(sPrefix)); 
    
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + sFrontEndType + ".ctl"],functionContent);  
  }
  
  
  
  
  // ------------------------------------------------------------------ STEP 4 : export ------------------------------------------------------------------ //   
  
            
  // --------------------------------------------------------- //
  // ----- File scripts/libs/TAG_prefixExportDevices.ctl ----- //
  // --------------------------------------------------------- //     
      
  string stringFileTAG_prefixExportDevices;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixExportDevices.ctl",stringFileTAG_prefixExportDevices);

    
  if (!mappingHasKey(mFileFunctions,"scripts/libs/" + sPrefix + "ExportDevices.ctl"))
  {
    // Mapping initialisation 
    mFileFunctions["scripts/libs/" + sPrefix + "ExportDevices.ctl"] = makeDynString();
    
    // Copy the file header (only the first time, because this file can be accessed multiple times during package generation)
    string header = substr(stringFileTAG_prefixExportDevices,0,strpos(stringFileTAG_prefixExportDevices,"//@{")) + "//@{\n\n";   
    strreplace(header,"TAG_prefix",sPrefix);
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "ExportDevices.ctl"],header);
  }

   // Extract function from the file
  string functionTAG_FrontEndType_Com_ExportConfig = extractFunctionFromStringFile(stringFileTAG_prefixExportDevices,"TAG_FrontEndType_Com_ExportConfig");
  strreplace(functionTAG_FrontEndType_Com_ExportConfig,"TAG_prefix",sPrefix);
  strreplace(functionTAG_FrontEndType_Com_ExportConfig,"TAG_PREFIX",strtoupper(sPrefix));
  strreplace(functionTAG_FrontEndType_Com_ExportConfig,"TAG_FrontEndType",sFrontEndType);
  strreplace(functionTAG_FrontEndType_Com_ExportConfig,"TAG_FRONTENDTYPE",strtoupper(sFrontEndType));
  dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "ExportDevices.ctl"],functionTAG_FrontEndType_Com_ExportConfig);
    
  
  
  
  // ------------------------------------------------------------------ STEP 5 : diagnostic panel ------------------------------------------------------------------ //   
    
    
  // -------------------------------------------------------------------------------------- //
  // ------ File panels/vision/graphicalFrame/TAG_FrontEndType_DiagnosticDetail.pnl ------- //
  // -------------------------------------------------------------------------------------- //   
  

  string stringFileTAG_FrontEndType_DiagnosticDetail;
  fileToString(sTemplateSourcePath + "/panels/vision/graphicalFrame/TAG_FrontEndType_DiagnosticDetail.pnl",stringFileTAG_FrontEndType_DiagnosticDetail);

  mFileFunctions["panels/vision/graphicalFrame/" + sFrontEndType + "_DiagnosticDetail.pnl"] = makeDynString(); 
  strreplace(stringFileTAG_FrontEndType_DiagnosticDetail,"TAG_FrontEndType",sFrontEndType);
  dynAppend(mFileFunctions["panels/vision/graphicalFrame/" + sFrontEndType + "_DiagnosticDetail.pnl"],stringFileTAG_FrontEndType_DiagnosticDetail);
  
  
  
  
  
  // ------------------------------------------------------------------ STEP 6 : front-end functions ------------------------------------------------------------------ //   
  // Done in step 3
  
  
  
  
  
  
  // ------------------------------------------------------------------ STEP 7 : icons, tree overview ------------------------------------------------------------------ //   
  
  
  // Just copy the icons provided in the template folder  
  copyFile(sTemplateSourcePath + "/pictures/TAG_FrontEndType/iconWithChildrenClose.bmp",          sNewPackagePath + "/" + sPackage + "/pictures/" + sFrontEndType + "/iconWithChildrenClose.bmp");
  int xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sFrontEndType + "/iconWithChildrenClose.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_FrontEndType/iconWithChildrenOpen.bmp",           sNewPackagePath + "/" + sPackage + "/pictures/" + sFrontEndType + "/iconWithChildrenOpen.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sFrontEndType + "/iconWithChildrenOpen.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_FrontEndType/iconWithChildrenPath.bmp",           sNewPackagePath + "/" + sPackage + "/pictures/" + sFrontEndType + "/iconWithChildrenPath.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sFrontEndType + "/iconWithChildrenPath.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_FrontEndType/iconWithChildrenSelected.bmp",       sNewPackagePath + "/" + sPackage + "/pictures/" + sFrontEndType + "/iconWithChildrenSelected.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sFrontEndType + "/iconWithChildrenSelected.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_FrontEndType/iconWithoutChildrenNotSelected.bmp", sNewPackagePath + "/" + sPackage + "/pictures/" + sFrontEndType + "/iconWithoutChildrenNotSelected.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sFrontEndType + "/iconWithoutChildrenNotSelected.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_FrontEndType/iconWithoutChildrenSelected.bmp",    sNewPackagePath + "/" + sPackage + "/pictures/" + sFrontEndType + "/iconWithoutChildrenSelected.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sFrontEndType + "/iconWithoutChildrenSelected.bmp");
  // function TAG_FrontEndType_ObjectListGetValueTime done in step 1
  
  
  
  
  
  
  
  // ------------------------------------------------------------------ STEP 8 : widget, catalog of widget, widget param ------------------------------------------------------------------ //  
        
  
  // ------------------------------------------------------------------------ //
  // ------ File panels/objects/FRONT_END/Widget_TAG_FrontEndType.pnl ------- //
  // ------------------------------------------------------------------------ //  
  
  string stringFileWidget_TAG_FrontEndType;
  fileToString(sTemplateSourcePath + "/panels/objects/FRONT_END/Widget_TAG_FrontEndType.pnl",stringFileWidget_TAG_FrontEndType);

  mFileFunctions["panels/objects/FRONT_END/Widget_" + sFrontEndType + ".pnl"] = makeDynString(); 
  // Warning, a line is "LANG:1 16 TAG_FrontEndType", where 16 is the size of the following string. This needs to be fixed
  strreplace(stringFileWidget_TAG_FrontEndType, "LANG:1 16 TAG_FrontEndType", "LANG:1 " + strlen(sFrontEndType) + " " + sFrontEndType);
  strreplace(stringFileWidget_TAG_FrontEndType,"TAG_prefix",sPrefix);
  strreplace(stringFileWidget_TAG_FrontEndType,"TAG_FrontEndType",sFrontEndType);
  dynAppend(mFileFunctions["panels/objects/FRONT_END/Widget_" + sFrontEndType + ".pnl"],stringFileWidget_TAG_FrontEndType);
    
  
  copyFile(sTemplateSourcePath + "/images/objects/FRONT_END/Widget_TAG_FrontEndType.pnl/Widget_TAG_FrontEndType.png",    
           sNewPackagePath +     "/" + sPackage + "/images/objects/FRONT_END/Widget_" + sFrontEndType + ".pnl/Widget_" + sFrontEndType + ".png");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./images/objects/FRONT_END/Widget_" + sFrontEndType + ".pnl/Widget_" + sFrontEndType + ".png");
    
  // -------------------------------------------------------------------------------------------- //
  // ------ File panels/objects_parampanels/FRONT_END/TAG_prefixTAG_FrontEndType_para.pnl ------- //
  // -------------------------------------------------------------------------------------------- //    

  file fileTAG_prefixTAG_FrontEndType_para = fopen(sTemplateSourcePath + "/panels/objects_parampanels/FRONT_END/TAG_prefixTAG_FrontEndType_para.pnl", "r");
  string stringFileTAG_prefixTAG_FrontEndType_para;
  fileToString(sTemplateSourcePath + "/panels/objects_parampanels/FRONT_END/TAG_prefixTAG_FrontEndType_para.pnl",stringFileTAG_prefixTAG_FrontEndType_para);
  fclose(fileTAG_prefixTAG_FrontEndType_para);

  mFileFunctions["panels/objects_parampanels/FRONT_END/" + sPrefix + sFrontEndType + "_para.pnl"] = makeDynString(); 
  strreplace(stringFileTAG_prefixTAG_FrontEndType_para,"TAG_FrontEndType",sFrontEndType);
  dynAppend(mFileFunctions["panels/objects_parampanels/FRONT_END/" + sPrefix + sFrontEndType + "_para.pnl"],stringFileTAG_prefixTAG_FrontEndType_para);
  
  
  
  
  
  
  
  
  
  
  
  // ------------------------------------------------------------------ STEP 9 : faceplate ------------------------------------------------------------------ //  
  
  
  // ------------------------------------------------------------------------------ //
  // ------ File panels/vision/graphicalFrame/FaceplateTAG_FrontEndType.pnl ------- //
  // ------------------------------------------------------------------------------ //     

  string stringFileFaceplateTAG_FrontEndType;
  fileToString(sTemplateSourcePath + "/panels/vision/graphicalFrame/FaceplateTAG_FrontEndType.pnl",stringFileFaceplateTAG_FrontEndType);

  mFileFunctions["panels/vision/graphicalFrame/Faceplate" + sFrontEndType + ".pnl"] = makeDynString(); 
  strreplace(stringFileFaceplateTAG_FrontEndType,"TAG_FrontEndType",sFrontEndType);
  dynAppend(mFileFunctions["panels/vision/graphicalFrame/Faceplate" + sFrontEndType + ".pnl"],stringFileTAG_prefixTAG_FrontEndType_para);
  
  // -------------------------------------------------------------------------------------- //
  // ------ File panels/vision/graphicalFrame/TAG_FrontEndType_DiagnosticDetail.pnl ------- //
  // -------------------------------------------------------------------------------------- //     

  file fileTAG_FrontEndType_DiagnosticDetail = fopen(sTemplateSourcePath + "/panels/vision/graphicalFrame/TAG_FrontEndType_DiagnosticDetail.pnl", "r");
  string stringFileTAG_FrontEndType_DiagnosticDetail;
  fileToString(sTemplateSourcePath + "/panels/vision/graphicalFrame/TAG_FrontEndType_DiagnosticDetail.pnl",stringFileTAG_FrontEndType_DiagnosticDetail);
  fclose(fileTAG_FrontEndType_DiagnosticDetail);

  mFileFunctions["panels/vision/graphicalFrame/" + sFrontEndType + "_DiagnosticDetail.pnl"] = makeDynString(); 
  strreplace(stringFileTAG_FrontEndType_DiagnosticDetail,"TAG_FrontEndType",sFrontEndType);
  dynAppend(mFileFunctions["panels/vision/graphicalFrame/" + sFrontEndType + "_DiagnosticDetail.pnl"],stringFileTAG_FrontEndType_DiagnosticDetail);
  
  
  
  
  
  
  
  
  
  // ------------------------------------------------------------------ STEP 10 : trending ------------------------------------------------------------------ //  
  

  // -------------------------------------------------------------- //
  // ---- File TAG_prefixTAG_FrontEndTypeTrendingFaceplate.pnl ---- //
  // -------------------------------------------------------------- //   
  // File missing... To complete when it is available  
  
  
  
  
  
  
  // ------------------------------------------------------------------ STEP 11 : system integrity ------------------------------------------------------------------ //  
  

  // -------------------------------------------------------------------------- //
  // ---- File scripts/libs/TAG_prefixSystemIntegrity_TAG_FrontEndType.ctl ---- //
  // -------------------------------------------------------------------------- // 
  
  string stringTAG_prefixSystemIntegrity_TAG_FrontEndType;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixSystemIntegrity_TAG_FrontEndType.ctl",stringTAG_prefixSystemIntegrity_TAG_FrontEndType);

  mFileFunctions["scripts/libs/" + sPrefix + "SystemIntegrity_" + sFrontEndType +".ctl"] = makeDynString();
  
  // File header
  string headerFileTAG_prefixSystemIntegrity_TAG_FrontEndType = 
      substr(
          stringTAG_prefixSystemIntegrity_TAG_FrontEndType,
          0,
          strpos(stringTAG_prefixSystemIntegrity_TAG_FrontEndType, "//@{")
      );   
  
  strreplace(headerFileTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefix",       sPrefix);
  strreplace(headerFileTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_FrontEndType", sFrontEndType);  
  
  // File global declarations
  string declarationFileTAG_prefixSystemIntegrity_TAG_FrontEndType = 
      substr(
        stringTAG_prefixSystemIntegrity_TAG_FrontEndType,
        strpos(stringTAG_prefixSystemIntegrity_TAG_FrontEndType,"//@{"),
        strpos(stringTAG_prefixSystemIntegrity_TAG_FrontEndType, "//@}") - strpos(stringTAG_prefixSystemIntegrity_TAG_FrontEndType, "//@{")
          )
      + "//@}\n\n";
  
  strreplace(declarationFileTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefix",       sPrefix);
  strreplace(declarationFileTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_FrontEndType", sFrontEndType);
  
  
  // Header and declaration are not separated
  dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "SystemIntegrity_" + sFrontEndType +".ctl"],
            headerFileTAG_prefixSystemIntegrity_TAG_FrontEndType
            + declarationFileTAG_prefixSystemIntegrity_TAG_FrontEndType + "//@{\n\n");
          
    
  
  // Replacement is proceeded each function at a time to have a better control of the process and possible future changes
  dyn_string functionListTAG_prefixSystemIntegrity_TAG_FrontEndType = makeDynString();
  dynAppend(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_FrontEndType_systemIntegrityInfo");
  dynAppend(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefixSystemIntegrity_TAG_FrontEndType_Initialize");
  dynAppend(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefixSystemIntegrity_TAG_FrontEndType_DataCallback");
  dynAppend(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefixSystemIntegrity_TAG_FrontEndType_HandleCommand");
  dynAppend(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefixSystemIntegrity_TAG_FrontEndType_checking");
  dynAppend(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType,"TAG_prefixSystemIntegrity_TAG_FrontEndTypeCheck");
  
  for (int i = 1 ; i <= dynlen(functionListTAG_prefixSystemIntegrity_TAG_FrontEndType) ; i++)
  {
    string functionContent = extractFunctionFromStringFile(stringTAG_prefixSystemIntegrity_TAG_FrontEndType,functionListTAG_prefixSystemIntegrity_TAG_FrontEndType[i]);
    // Tag replacement
    strreplace(functionContent,"TAG_prefix",sPrefix); 
    strreplace(functionContent,"TAG_FrontEndType",sFrontEndType); 
    strreplace(functionContent,"TAG_FRONTENDTYPE",strtoupper(sFrontEndType)); 
    strreplace(functionContent,"TAG_PREFIX",strtoupper(sPrefix)); 
    
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "SystemIntegrity_" + sFrontEndType +".ctl"],functionContent);  
  }
  
    
  // ---------------------------------------------------------------------------------------------------- //
  // ------ File panels/vision/systemIntegrity/systemIntegrity_TAG_FrontEndType_configuration.pnl ------- //
  // ---------------------------------------------------------------------------------------------------- //     

  string stringFilesystemIntegrity_TAG_FrontEndType_configuration;
  fileToString(sTemplateSourcePath + "/panels/vision/systemIntegrity/systemIntegrity_TAG_FrontEndType_configuration.pnl",stringFilesystemIntegrity_TAG_FrontEndType_configuration);

  mFileFunctions["panels/vision/systemIntegrity/systemIntegrity_" + sFrontEndType + "_configuration.pnl"] = makeDynString(); 
  // Warning, a line is "LANG:1 25 TAG_FrontEndType time out", where 25 is the size of the following string. This needs to be fixed
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_configuration, "LANG:1 25 TAG_FrontEndType time out", "LANG:1 " + strlen(sFrontEndType + " time out") + " " + sFrontEndType + " time out");
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_configuration,"TAG_prefix",sPrefix); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_configuration,"TAG_FrontEndType",sFrontEndType); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_configuration,"TAG_FRONTENDTYPE",strtoupper(sFrontEndType)); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_configuration,"TAG_PREFIX",strtoupper(sPrefix)); 
  dynAppend(mFileFunctions["panels/vision/systemIntegrity/systemIntegrity_" + sFrontEndType + "_configuration.pnl"],stringFilesystemIntegrity_TAG_FrontEndType_configuration);
  
    
  // ---------------------------------------------------------------------------------------------------- //
  // -------- File panels/vision/systemIntegrity/systemIntegrity_TAG_FrontEndType_operation.pnl --------- //
  // ---------------------------------------------------------------------------------------------------- //     

  string stringFilesystemIntegrity_TAG_FrontEndType_operation;
  fileToString(sTemplateSourcePath + "/panels/vision/systemIntegrity/systemIntegrity_TAG_FrontEndType_operation.pnl",stringFilesystemIntegrity_TAG_FrontEndType_operation);
  
  mFileFunctions["panels/vision/systemIntegrity/systemIntegrity_" + sFrontEndType + "_operation.pnl"] = makeDynString(); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_operation,"TAG_prefix",sPrefix); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_operation,"TAG_FrontEndType",sFrontEndType); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_operation,"TAG_FRONTENDTYPE",strtoupper(sFrontEndType)); 
  strreplace(stringFilesystemIntegrity_TAG_FrontEndType_operation,"TAG_PREFIX",strtoupper(sPrefix)); 
  dynAppend(mFileFunctions["panels/vision/systemIntegrity/systemIntegrity_" + sFrontEndType + "_operation.pnl"],stringFilesystemIntegrity_TAG_FrontEndType_operation);
  
  
  
  // ------------------------------------------------------------------------------- //
  // -------- File msg/en_US.iso88591/TAG_FrontEndType_systemIntegrity.cat --------- //
  // ------------------------------------------------------------------------------- //    
  string stringFileTAG_FrontEndType_systemIntegrity;
  fileToString(sTemplateSourcePath + "/msg/en_US.iso88591/TAG_FrontEndType_systemIntegrity.cat",stringFileTAG_FrontEndType_systemIntegrity);
  mFileFunctions["msg/en_US.iso88591/" + sFrontEndType + "_systemIntegrity.cat"] = makeDynString(); 
  strreplace(stringFileTAG_FrontEndType_systemIntegrity,"TAG_FrontEndType",sFrontEndType);
  dynAppend(mFileFunctions["msg/en_US.iso88591/" + sFrontEndType + "_systemIntegrity.cat"],stringFileTAG_FrontEndType_systemIntegrity);
  
  
  
  
  
  
  
  
  
  
  
  
  // The front-end now exists, the appropriate flag must be changed so that it is not created twice
  
  // Changing in the front-end list used for generation
  int frontEndPos = dynContains(getDynString(frontEndTypeList,LIST_NAME_INDEX), sFrontEndType);
  if (frontEndPos > 0)
  {
    frontEndTypeList[frontEndPos][LIST_EXISTS_INDEX] = true;
  }
  
  // Changing in the table  
  for (int i = 0 ; i < tableFrontEndList.lineCount ; i++)
  {
    string currentFrontEndName;
    getValue("tableFrontEndList","cellValueRC",i,COLUMN_FRONT_END_NAME,currentFrontEndName);  
    if (currentFrontEndName == sFrontEndType)
    {
      setValue("tableFrontEndList","cellValueRC",i,COLUMN_FRONT_END_EXISTS,true);  
    }
  } 
}


//@}
