/**@name LIBRARY: unMessageText.ctl

@authors: Vincent Forest , Herve Milcent (LHC-IAS)

Creation Date: 23/07/2002
Modification History: None
		02/07/2004: Herve 
			- in function unMessageText_getDpMessage: get the local unMessageText DP instead of all of them.
			set UN_MESSAGE_TEXT_HISTORY_PERIOD to 1800 sec (30min)
			- in function unMessageText_send: always sent to the local unMessageText DP.

			
version 1.0

External Functions: 
	- unMessageText_createDpMessage: create the DP _unMessageText in the DP type _UnMessageText
	- unMessageText_send: send messages throw the distributed system
	- unMessageText_sendException: send exceptionInfo throw the distributed system
	- unMessageText_getDpMessage: get the DP name that contains the message to display

Purpose: 
This library contain generic functions to be use with the graphical frame component 
UNICOS\PVSS\panels\objects\UN_INFOS\unMessageText.pnl

Usage: Public

PVSS manager usage: UI (PVSS00NV) and UI extended (PVSS00NG)

Constraints:
		. a DP type _UnMessageText must be existing
		. PVSS version: 3.0 
		. operating system: NT and W2000, but tested only under W2000.
		. distributed system: yes.
*/


// Global constants for Message Text component
const string UN_MESSAGE_TEXT_DELIMITER = "\t";			// Message text delimiter
const string UN_MESSAGE_TEXT_TYPE_DELIMITER = "+";		// Message text type delimiter
const string UN_MESSAGE_TEXT_HISTORY_DELIMITER = "~"; 	// Delimiter for message history
const string UN_MESSAGE_TEXT_LOGIN_DELIMITER = "="; 	// Delimiter for login field
const unsigned UN_MESSAGE_TEXT_FIELD_NUMBER = 6;		// Filter numbers
const unsigned UN_MESSAGE_TEXT_FILTER_RECEIVER_SYSTEM_NAME = 1;	// Message text receiver system name filter position in the filter configuration
const unsigned UN_MESSAGE_TEXT_FILTER_RECEIVER_UI_NUMBER = 2;	// Message text receiver ui number filter position in the filter configuration
const unsigned UN_MESSAGE_TEXT_FILTER_SENDER = 3;		// Message text sender filter position in the filter configuration
const unsigned UN_MESSAGE_TEXT_FILTER_TYPE = 4;			// Message text type filter position in the filter configuration
const unsigned UN_MESSAGE_TEXT_FILTER_TEXT = 5;			// Message text text filter position in the filter configuration
const unsigned UN_MESSAGE_TEXT_FIELD_LOGIN = 6;			// Message text login field
const unsigned UN_MESSAGE_TEXT_HISTORY_PERIOD = 1800;	// In seconds, the default display period


// ------------------------------------------------------------------------------------------------------------------------------------
//@{

// unMessageText_createDpMessage
/**
Purpose:	create the DP _unMessageText in the DP type _UnMessageText

Usage: 		Public

PVSS manager usage: UI (PVSS00NV) and UI extended (PVSS00NG)

Parameter:	dyn_string& exceptionInfo, used in the error handling process (fwException_raise and fwExceptionHandling_display functions)

Constraints:
		. a DP type _UnMessageText must be existing
		. PVSS version: 3.0 
		. operating system: NT and W2000, but tested only under W2000.
		. distributed system: yes.
*/
void unMessageText_createDpMessage(dyn_string& exceptionInfo)
{
	dyn_string exceptionInfoTemp;
	
	// Check if _unMessageText exists and creates it if necessary
	if (unMessageText_getDpMessage(exceptionInfoTemp) == "") 
		if (dpCreate("_unMessageText","_UnMessageText") == -1)	// Create DP
			fwException_raise(exceptionInfo,"ERROR","Library unMessageText.ctl : can't create DP _unMessageText","");
}

// ------------------------------------------------------------------------------------------------------------------------------------

// unMessageText_send
/**
Purpose:	send messages throw the distributed system

Usage: 		Public

PVSS manager usage: UI (PVSS00NV) and UI extended (PVSS00NG)

Parameters:	string messageToSystemName, receiver system name, * = every system name
			string messageToUiNumber, receiver ui number, * = every ui number			
			string messageSender, sender identifier
			string userOrGroup, for access control, equal to "user" or "group"
			string userOrGroupName, for access control, the user or group name ex: "root"
			string messageType, message type : ERROR, WARNING, INFO or EXPERTINFO
			string messageText, the message itself
			
			dyn_string& exceptionInfo, used in the error handling process (fwException_raise and fwExceptionHandling_display functions)

Constraints:
		. PVSS version: 3.0 
		. operating system: NT and W2000, but tested only under W2000.
		. distributed system: yes.
*/

void unMessageText_send(string messageToSystemName, string messageToUiNumber, string messageSender, 
						string userOrGroup, string userOrGroupName,
				   		string messageType, string messageText, dyn_string& exceptionInfo)
{
	string dpMessage;								// Global DP that contains messages
	bool sendError = false;
	
	if ((messageToSystemName == "") || (messageToUiNumber == "") || (messageSender == "") || (messageText == "") || (userOrGroupName == ""))
		{
		fwException_raise(exceptionInfo, "ERROR", "unMessageText Component : bad message format.","");
		sendError = true;
		}
	if ((messageType != "INFO") && (messageType != "EXPERTINFO") && (messageType != "ERROR") && (messageType != "WARNING"))
		{
		fwException_raise(exceptionInfo, "ERROR", "unMessageText Component : bad message type.","");
		sendError = true;
		}
	if ((userOrGroup != "user") && (userOrGroup != "group"))
		{
		fwException_raise(exceptionInfo, "ERROR", "unMessageText Component : bad message type.","");
		sendError = true;
		}
	
	if (sendError == false)
	{
		dpMessage = unMessageText_getDpMessage(exceptionInfo);	// Get DP name
		if (dpMessage != "")						// Send message
		{
			dpMessage = dpMessage + ":_original.._value";
			if (dpSet(dpMessage, messageToSystemName + UN_MESSAGE_TEXT_DELIMITER + messageToUiNumber + UN_MESSAGE_TEXT_DELIMITER +
							 	 messageSender + UN_MESSAGE_TEXT_DELIMITER +
							 	 messageType + UN_MESSAGE_TEXT_DELIMITER + messageText + UN_MESSAGE_TEXT_DELIMITER +
							 	 userOrGroup + UN_MESSAGE_TEXT_LOGIN_DELIMITER + userOrGroupName) == -1)
			{
				fwException_raise(exceptionInfo, "ERROR", "Library unMessageText.ctl : dpSet failed. Message wasn't sent","");
				sendError = true;
			}
		}
		else
		{
			fwException_raise(exceptionInfo, "ERROR", "Library unMessageText.ctl : no unMessageText DP","");
			sendError = true;
		}
	}
		
	if (sendError == true)
		{
		fwException_raise(exceptionInfo, "ERROR", 
						  "unMessageText_send : error trying to send message=" + messageText + ", type=" + messageType +
						  ", toSystemName=" + messageToSystemName + ", toUiNumber=" + messageToUiNumber +
						  ", to " + userOrGroup + "=" + userOrGroupName +
						  ", from sender=" + messageSender,"");
		}
}
// ------------------------------------------------------------------------------------------------------------------------------------

// unMessageText_getDpMessage
/**
Purpose:	get the DP name that contains the message to display

Usage: 		Public

PVSS manager usage: UI (PVSS00NV) and UI extended (PVSS00NG)

Parameters:	dyn_string& exceptionInfo, used in the error handling process (fwException_raise and fwExceptionHandling_display functions)

Constraints:
		. Only one DP named "*:_unMessageText.message:" must be existing in the distributed system
		. PVSS version: 3.0 
		. operating system: NT and W2000, but tested only under W2000.
		. distributed system: yes.
*/
string unMessageText_getDpMessage(dyn_string& exceptionInfo)
{
	dyn_string dpMessageTextNamesArray;			// Array containing each DP named [*]:_unMessageText.message in the DPT _UnMessageText
	int namesArrayLength;						// dpMessageTextNamesArray length
	
	// Get the DP name that contains the message text
	// Its name format is [*]:_unMessageText.message
	// Steps : get each DPE from this format, check for errors
	dpMessageTextNamesArray = dpNames(getSystemName()+"_unMessageText.message","_UnMessageText");
	namesArrayLength = dynlen(dpMessageTextNamesArray);
	if (namesArrayLength <= 0)
		{
		fwException_raise(exceptionInfo, "ERROR", "Library unMessageText.ctl : the data point [*]:_unMessageText.message is not existing","");
		return("");
		}
	else 
		return(dpMessageTextNamesArray[1]);
}

// ------------------------------------------------------------------------------------------------------------------------------------

// unMessageText_sendException
/**
Purpose:	send exceptionInfo throw the distributed system

Usage: 		Public

PVSS manager usage: UI (PVSS00NV) and UI extended (PVSS00NG)

Parameters:	string messageToSystemName, receiver system name, * = every system name
			string messageToUiNumber, receiver ui number, * = every ui number			
			string messageSender, sender identifier
			string userOrGroup, for access control, equal to "user" or "group"
			string userOrGroupName, for access control, the user or group name ex: "root"
			dyn_string &exceptionInfo, contains messageType and messageText

Constraints:
		. Only one DP named "*:_unMessageText.message" must be existing in the distributed system
		. PVSS version: 3.0 
		. operating system: NT and W2000, but tested only under W2000.
		. distributed system: yes.
*/
void unMessageText_sendException(string messageToSystemName, string messageToUiNumber, string messageSender, 
						string userOrGroup, string userOrGroupName,	dyn_string &exceptionInfo)
{
	int numberOfMessages;
	int i;
	dyn_string errorArray;
	
	numberOfMessages = dynlen(exceptionInfo);
	for(i=1; i<=(numberOfMessages/3); i++)
	{
		unMessageText_send(messageToSystemName, messageToUiNumber, messageSender, userOrGroup, userOrGroupName,
				   		   exceptionInfo[3*i-2], exceptionInfo[3*i-1], errorArray);
	}
	exceptionInfo = makeDynString();
	if (dynlen(errorArray) > 0)
		dynAppend(exceptionInfo, errorArray);
}
//@}